


/**********************************************************************************/
/* This function implements level set image segmentation segmentation by minimizing
   a kernel-induced data term and a systematic general mapping between the segmentation 
   regions and the regions defined by the curves and their intersections. 

   References:

   M. Ben Salah, A. Mitiche and I. Ben Ayed, Effective Level Set Image Segmentation
   with a Kernel Induced Data Term, IEEE Transactions on Image processing,  vol. 19, no 1, 
   pp. 220-232, 2010. 

   A. Mansouri, A. Mitiche, and C. Vazquez, Multiregion competition: A level set extension 
   of region competition to multiple region partioning,  Computer Vision and Image Understanding,
   vol. 101, no. 3, pp. 137150, 2006.

  
/**********************************************************************************/



#include "levelset.h"
#include <math.h>

 /**********************************************************************************
                    Compute the data-driven velocity
************************************************************************************/
float *Seg_Part_Ker(List_LevelSet *p_lls, Image *p_image, int p, float *F)
{

  int k,l;
  float uu,vv,I,uu_m,vv_m;
  
  I = p_image->I0[p];
  
  for(k=0;k<p_lls->Num;k++){
    
    vv_m = p_lls->p_param_out[2];
    vv   = (-exp(-(I-vv_m)*(I-vv_m))/sigma);
    uu_m = p_lls->l_levelset[k].p_param[2]; 
    uu   = -exp(-(I-uu_m)*(I-uu_m)/sigma);    
    F[k] = 10000*(uu-vv);

    for(l=0;l<p_lls->Num;l++){
      if((p_lls->l_levelset[l].U[p]>0.0)||
	 (p_lls->l_levelset[l].U[p-1]>0.0)||
	 (p_lls->l_levelset[l].U[p+1]>0.0)||
	 (p_lls->l_levelset[l].U[p-p_image->xSize]>0.0)||
	 (p_lls->l_levelset[l].U[p+p_image->xSize]>0.0)){
	if(l<k){
	  F[k] = 0.0;
	  break;
	}
	if(l==k)
	  continue;
	vv_m = p_lls->l_levelset[l].p_param[2];
	vv   = -exp(-(I-vv_m)*(I-vv_m)/sigma);
	F[k] = 10000*(uu-vv);
	break;
      }
    }
  }
  return(F);
  
}


/***************************************************************************/
/* Calculate the mean and the area of domains inside and outside the curve */
/***************************************************************************/
int Init_Part_Ker(List_LevelSet *p_lls, Image *p_image)
{
  const int Num_Param = 3;  
  int v,h,offset,i;
  int flag;
  
  for(i=0;i<p_lls->Num;i++){
    p_lls->l_levelset[i].Par_Num = Num_Param;
    if((p_lls->l_levelset[i].p_param = 
	(double*)calloc(Num_Param,sizeof(double)))==NULL)
      return(IMG_FATAL);
  }
  if((p_lls->p_param_out = (double*)calloc(Num_Param,sizeof(double)))==NULL)
      return(IMG_FATAL);

  flag = IMG_FALSE;
  for(v=0;v<p_image->ySize;v++) {
    for(h=0;h<p_image->xSize;h++) {
      offset = v*p_image->xSize+h;
      flag = IMG_FALSE;
      for(i=0;i<p_lls->Num;i++){
	if (p_lls->l_levelset[i].U[offset] > 0.0) {
	  p_lls->l_levelset[i].p_param[0] += 1.0;
	  p_lls->l_levelset[i].p_param[1] += (double)p_image->I0[offset];
	  flag = IMG_TRUE;
	  break;
	}
      }
      if(!flag){
	p_lls->p_param_out[0] += 1.0;
	p_lls->p_param_out[1] += (double)p_image->I0[offset];
      }      
    }
  }
  for(i=0;i<p_lls->Num;i++){
    p_lls->l_levelset[i].p_param[2] = 
      p_lls->l_levelset[i].p_param[1]/p_lls->l_levelset[i].p_param[0];
  }
  p_lls->p_param_out[2] = p_lls->p_param_out[1]/p_lls->p_param_out[0];
  
  return(IMG_NO_ERROR);
  
}

/*****************************************/
/*        Update parametres              */
/*****************************************/
int Update_Part_Ker(List_LevelSet *p_lls, Image *p_image,
		   int p){

  double error;
  int k,l;
  unsigned char flag_change = IMG_FALSE;
  unsigned char flag_finish = IMG_FALSE;

  k = 0;
  while((!flag_finish)&&(k<p_lls->Num)){ 
    
    if(p_lls->l_levelset[k].U[p]*p_lls->l_levelset[k].Un[p] < 0.0){
      
      flag_change = IMG_TRUE;
      error = ((double)p_image->I0[p]-p_lls->l_levelset[k].p_param[2])*
      ((double)p_image->I0[p]-p_lls->l_levelset[k].p_param[2])/sigma;
      if(p_lls->l_levelset[k].Un[p]>0.0){
	
	p_lls->l_levelset[k].p_param[0] += exp(-error);
	p_lls->l_levelset[k].p_param[1] += (double)p_image->I0[p]*exp(-error);
	if(p_lls->l_levelset[k].p_param[0]==0)
	  p_lls->l_levelset[k].p_param[2] = 0;
	else
	  p_lls->l_levelset[k].p_param[2] =
	    p_lls->l_levelset[k].p_param[1]/p_lls->l_levelset[k].p_param[0];   
	
	l = k+1;
	while((!flag_finish)&&(l<p_lls->Num)){
	  if(p_lls->l_levelset[l].U[p]>0.0){
	   
            error = ((double)p_image->I0[p]-p_lls->l_levelset[l].p_param[2])*
            ((double)p_image->I0[p]-p_lls->l_levelset[l].p_param[2])/sigma;
	    p_lls->l_levelset[l].p_param[0] -= exp(-error);
	    p_lls->l_levelset[l].p_param[1] -= (double)p_image->I0[p]*exp(-error);
	    if(p_lls->l_levelset[l].p_param[0]==0)
	      p_lls->l_levelset[l].p_param[2] = 0;
	    else
	      p_lls->l_levelset[l].p_param[2] =
		p_lls->l_levelset[l].p_param[1]/p_lls->l_levelset[l].p_param[0];   
	    flag_finish = IMG_TRUE;
	  }
	  l++;
	}
	if(!flag_finish){
	
          error = ((double)p_image->I0[p]-p_lls->p_param_out[2])*
          ((double)p_image->I0[p]-p_lls->p_param_out[2])/sigma;
	  p_lls->p_param_out[0] -= exp(-error);
	  p_lls->p_param_out[1] -= (double)p_image->I0[p]*exp(-error);
	  if(p_lls->p_param_out[0]==0)
	    p_lls->p_param_out[2] = 0;
	  else
	    p_lls->p_param_out[2] =
	      p_lls->p_param_out[1]/p_lls->p_param_out[0];   
	  flag_finish = IMG_TRUE;
	}
      }
      else{
	
        error = ((double)p_image->I0[p]-p_lls->l_levelset[k].p_param[2])*
        ((double)p_image->I0[p]-p_lls->l_levelset[k].p_param[2])/sigma; 
	p_lls->l_levelset[k].p_param[0] -= exp(-error);
	p_lls->l_levelset[k].p_param[1] -= (double)p_image->I0[p]*exp(-error);
	if(p_lls->l_levelset[k].p_param[0]==0)
	  p_lls->l_levelset[k].p_param[2] = 0;
	else
	  p_lls->l_levelset[k].p_param[2] =
	    p_lls->l_levelset[k].p_param[1]/p_lls->l_levelset[k].p_param[0];   
	l = k+1;
	while((!flag_finish)&&(l<p_lls->Num)){
	  if(p_lls->l_levelset[l].Un[p]>0.0){
            error = ((double)p_image->I0[p]-p_lls->l_levelset[l].p_param[2])*
            ((double)p_image->I0[p]-p_lls->l_levelset[l].p_param[2])/sigma;
	    p_lls->l_levelset[l].p_param[0] += exp(-error);
	    p_lls->l_levelset[l].p_param[1] += (double)p_image->I0[p]*exp(-error);
	    if(p_lls->l_levelset[l].p_param[0]==0)
	      p_lls->l_levelset[l].p_param[2] = 0;
	    else
	      p_lls->l_levelset[l].p_param[2] =
		p_lls->l_levelset[l].p_param[1]/p_lls->l_levelset[l].p_param[0];   
	    flag_finish = IMG_TRUE;
	  }
	  l++;
	}
	if(!flag_finish){
          error = ((double)p_image->I0[p]-p_lls->p_param_out[2])*
          ((double)p_image->I0[p]-p_lls->p_param_out[2])/sigma;
	  p_lls->p_param_out[0] += exp(-error);
	  p_lls->p_param_out[1] += (double)p_image->I0[p]*exp(-error);
	  if(p_lls->p_param_out[0]==0)
	    p_lls->p_param_out[2] = 0;
	  else
	    p_lls->p_param_out[2] =
	      p_lls->p_param_out[1]/p_lls->p_param_out[0];   
	  flag_finish = IMG_TRUE;
	}
      }
    }
    else{
      
      if((p_lls->l_levelset[k].U[p] > 0.0)&&(p_lls->l_levelset[k].Un[p] > 0.0)){
	break;
      }
    }
    k++;
  }
 
  return(flag_change); 
}
